/**
 * @fileoverview This rule enforces to use appropriate i18n keys instead of hardcoding content
 * @author Aravindh T
 */
let methods = []
let props = []
let objectKeys = []
let allowedTags = []
let keys = []
let commentsToIgnoreHardcodingCheck = []
let stringsAllowedAsBlockCommentStart = []
let stringsAllowedAsBlockCommentEnd = []
let dateFormatPatterns = []
let allowedStringStartWords = []
let allowedPropStartingWords = []
let allowedPropEndingWords = []
let validTagsLocationObject = null
let aliasNamesOfgetI18NValueMethod = []
let aliasNamesOfExcludedMethods = []
let allowedWordsThatCanBePresentInAComponentNameToExcludeTheComponent = []

//regexp patterns 
const pattern = /(\w\s)+((&|,)\s)*\w+(\s\w)*/ //   /(\w\s)+\w+/
const cssStylingPattern= /([0-9]+(px|%|em|rem|vh|vw)(\s|$))+(\w+(\s|$))*(#[a-zA-Z0-9]+)?/
const sequenceOfNumbers = /([0-9]+\s)+/
const numberPattern = /[0-9]+/
const booleanPattern = /true|false/
const keyPattern = /\w+(\.\w)+/
const camelCasePattern = /^[a-z]+[A-Z]+/
const PascalCasePattern = /^[A-Z]{1}[a-z]+[A-Z]{1}[a-z]+/
const symbolPattern = /[&|\-|#|,|.|?|:|*|<|=|>|(|)|\/|\\|{|}|[|\]]/

const ERROR_MESSAGE = 'Do not hardcode content. Use I18N key instead' //no i18n
const ERROR_MESSAGE_JSX_TEXT = "Do not hardcode string as JSX children in the JSX tag. Use i18n key instead"

module.exports={
    meta:{
        type: "problem",
        docs: {
            description: "This rule prevents hardcoding content and enforces to use appropriate i18n keys" //no i18n
        }

    },
    create: function(context){
        let stringInitializedIdentifiers = []
        let stringInitObjectProperties = []
        const excludes = context.options[0]
        methods = excludes.methods || []
        props = excludes.props || []
        objectKeys = excludes.objectKeys || []
        allowedTags = excludes.allowedTags || []
        keys = excludes.keys || []
        commentsToIgnoreHardcodingCheck = excludes.commentsToIgnoreHardcodingCheck || []
        stringsAllowedAsBlockCommentStart = excludes.stringsAllowedAsBlockCommentStart || []
        stringsAllowedAsBlockCommentEnd = excludes.stringsAllowedAsBlockCommentEnd || []
        dateFormatPatterns = excludes.dateFormatPatterns || []
        allowedStringStartWords = excludes.startString || []
        allowedPropStartingWords = excludes.allowedPropStartingWords || []
        allowedPropEndingWords = excludes.allowedPropEndingWords || []
        allowedWordsThatCanBePresentInAComponentNameToExcludeTheComponent = excludes.allowedWordsThatCanBePresentInAComponentNameToExcludeTheComponent || []

        return{
            Literal: function(node){
                //boolean variables for reporting errors
                let literalText = node.value
                let isHardCoded = false
                let isSingleWordHardcoded = false
                let isASequenceOfNumbers = false 
                let isAStringInACondition = false 
                let isAStringInAllowedAttribute = false 
                let isI18NNotNeeded = false 
                let isAStringInAllowedProp = false 
                let isAStringInAllowedTag = false 
                let isACSSProperty = false 
                let isAPropertyName  = false 
                let isAMemberExpression = false 
                let isValidKeyInsideValidMethod = false 
                let isAStringInAllowedMethod = false 
                let isAStringInsideIgnoredBlock = false 
                let isAStringInAttributeOfAllowedTag = false 
                let isAStringInNewExpression = false  
                let isAStringInThrowStatement = false 
                let isADateFormatPattern = false 
                let isACamelCasedString = false 

                let isANumber = false 
                let isABooleanValue = false 
                let isAKeyPattern = false 
                let isAStringInImportStatement = false 
                let isSingleWordStringPassedAsArgument = false 
                let isUnderscorePresentInString = false 
                let isAnAllowedString = false 
                let isARegexPattern = false 
                let isAnEmptyString = false 
                let isAPascalCasedString = false 
                let isASingleSymbol = false 
                let isSingleWordInitAsAssignment = false
                let isSingleWordInitAsObjectProperty = false
                let isSingleWordInitInArray = false
                let isAStringInExportDeclaration = false

                //ancestors of literal node
                let ancestorsOfLiteral = context.getAncestors();
                let immediateAncestorOfLiteral = ancestorsOfLiteral[ancestorsOfLiteral.length -1]
                let grandParentOfLiteral = ancestorsOfLiteral[ancestorsOfLiteral.length -2]
                let greatGrandParentOfLiteral = ancestorsOfLiteral[ancestorsOfLiteral.length -3]

                let stringStartsAt = node.loc.start.line
                let commentObj = {}
                
                isHardCoded = pattern.test(literalText)
                isSingleWordHardcoded = literalText && !isHardCoded
                
                //checking comments
                if(isHardCoded || isSingleWordHardcoded){
                    commentObj = isCommentAdded(ancestorsOfLiteral,stringStartsAt,immediateAncestorOfLiteral)
                    isI18NNotNeeded = commentObj.isCommentAdded
                }
                //checking block comments
                if(isHardCoded || isSingleWordHardcoded){
                    isAStringInsideIgnoredBlock = commentObj.isBlockCommentAdded
                }

                //checking if the string is a Regex pattern
                isARegexPattern = node.regex ? true : false

                //checking if string is a part of an allowed tag
                isAStringInAllowedTag = isHardcodedInAllowedTag(stringStartsAt)

                //checking if a string is in an attribute of allowed tags
                isAStringInAttributeOfAllowedTag = isHardcodedInAttributeOfAllowedTag(ancestorsOfLiteral)

                //checking if string is part of a condition
                isAStringInACondition = isHardcodedInCondition(immediateAncestorOfLiteral)
                
                //checking if string is an allowed key/property
                isAStringInAllowedAttribute = isHardcodedInAllowedProperty(immediateAncestorOfLiteral,grandParentOfLiteral)
                
                //checking if string is in an allowed prop
                if(immediateAncestorOfLiteral.type === 'JSXAttribute'){
                    isAStringInAllowedProp = isHardcodedInAllowedProp(immediateAncestorOfLiteral)
                }
                else if(grandParentOfLiteral.type === 'JSXAttribute'){
                    isAStringInAllowedProp = isHardcodedInAllowedProp(grandParentOfLiteral)
                }
                else if(immediateAncestorOfLiteral.type === 'LogicalExpression'
                     && grandParentOfLiteral.type === 'JSXExpressionContainer'
                     && greatGrandParentOfLiteral.type === 'JSXAttribute'){
                        isAStringInAllowedProp = isHardcodedInAllowedProp(greatGrandParentOfLiteral)
                }

                //checking for css styling pattern
                isACSSProperty = isAHardcodedCSSProperty(literalText)

                //checking if string is a date format pattern
                isADateFormatPattern = isAHardcodedDateFormatPattern(literalText)

                //checking if string is in a throw statement
                isAStringInThrowStatement = isHardcodedInThrowStmt(immediateAncestorOfLiteral)
                
                //checking if string is in new expression
                isAStringInNewExpression = isHardcodedInNewExpression(immediateAncestorOfLiteral)

                //checking if the strings is a property names
                isAPropertyName = isHardCodedInObjectKey(immediateAncestorOfLiteral)
                
                //omitting strings that are MemberExpression - obj['key'] - MemberExpression example
                isAMemberExpression = isHardcodedAsMemberExpression(immediateAncestorOfLiteral)

                //checking if the string is a valid key and is used inside getI18NValue method
                isValidKeyInsideValidMethod = isValidKeyInValidMethod(immediateAncestorOfLiteral,literalText)
                
                //checking if the more than one word string is hardcoded inside allowed method 
                if(isHardCoded){
                    if(immediateAncestorOfLiteral.type === 'CallExpression'){
                        isAStringInAllowedMethod = isHardcodedInAllowedMethods(immediateAncestorOfLiteral)
                    }
                    else if(immediateAncestorOfLiteral.type === 'Property'
                            && grandParentOfLiteral.type === 'ObjectExpression'
                            && greatGrandParentOfLiteral.type === 'CallExpression'){
                                isAStringInAllowedMethod = isHardcodedInAllowedMethods(greatGrandParentOfLiteral)
                    }
                }

                //checking if the string is a sequence of numbers
                isASequenceOfNumbers = isASequenceOfNumbersHardcoded(literalText)

                //the following checks are only for single word hardcoding
                //checking if single word hardcoded string is passed as argument to a function
                if(isSingleWordHardcoded){

                    //checking if string is in import statement
                    isAStringInImportStatement = immediateAncestorOfLiteral.type === 'ImportDeclaration' ? true : false

                    //checking if a string is in an export statement
                    isAStringInExportDeclaration = immediateAncestorOfLiteral.type === 'ExportNamedDeclaration' 
                                                || immediateAncestorOfLiteral.type === 'ExportAllDeclaration' ? true : false

                    //checking if single word string is passed as function argument
                    isSingleWordStringPassedAsArgument = immediateAncestorOfLiteral.type === 'CallExpression'
                                                        || grandParentOfLiteral.type === 'CallExpression' ? true : false
                    
                    // //checking if a function argument is init with string
                    if(!isSingleWordStringPassedAsArgument){
                        isSingleWordStringPassedAsArgument = immediateAncestorOfLiteral.type === 'AssignmentPattern'
                                                            && grandParentOfLiteral.type === 'FunctionExpression'
                                                            && greatGrandParentOfLiteral.type === 'MethodDefinition'
                                                            ? true : false
                    }

                    //checking if an incorrect single word is passed as argument to getI18NValue method
                    if(isSingleWordStringPassedAsArgument){
                        if(immediateAncestorOfLiteral.type === 'CallExpression' && immediateAncestorOfLiteral.callee){
                            if(immediateAncestorOfLiteral.callee.type === 'MemberExpression'){
                                if(immediateAncestorOfLiteral.callee.property.name === 'getI18NValue'
                                || aliasNamesOfgetI18NValueMethod.includes(immediateAncestorOfLiteral.callee.property.name)
                                ){
                                    isSingleWordStringPassedAsArgument = isValidKeyInValidMethod(immediateAncestorOfLiteral,literalText) ? true : false
                                }
                            }
                            else if(immediateAncestorOfLiteral.callee.type === 'Identifier'){
                                if(immediateAncestorOfLiteral.callee.name === 'getI18NValue'
                                || aliasNamesOfgetI18NValueMethod.includes(immediateAncestorOfLiteral.callee.name)
                                ){
                                    isSingleWordStringPassedAsArgument = isValidKeyInValidMethod(immediateAncestorOfLiteral,literalText) ? true : false
                                }
                            }
                        }
                    }

                    //checking if a string is in camelCase
                    isACamelCasedString = camelCasePattern.test(literalText)

                    //checking if string is initialized as variable assignment
                    isSingleWordInitAsAssignment = immediateAncestorOfLiteral.type === 'VariableDeclarator'
                                                    || immediateAncestorOfLiteral.type === 'AssignmentExpression' ? true : false
                    if(!isSingleWordInitAsAssignment){
                        isSingleWordInitAsAssignment = immediateAncestorOfLiteral.type === 'LogicalExpression'
                                                        && (grandParentOfLiteral.type === 'VariableDeclarator' ||
                                                        grandParentOfLiteral.type === 'AssignmentExpression') ? true : false
                    }
                    
                    //checking if string is initialized as object property value
                    isSingleWordInitAsObjectProperty = immediateAncestorOfLiteral.type === 'Property'
                                                        || grandParentOfLiteral.type === 'Property' ? true : false

                    //checking if string is initialized inside an array
                    isSingleWordInitInArray = immediateAncestorOfLiteral.type === 'ArrayExpression' ? true : false

                    //checking if string is a number or a boolean value or a key format 
                    isANumber = numberPattern.test(literalText)
                    isABooleanValue = booleanPattern.test(literalText)
                    isAKeyPattern = keyPattern.test(literalText)

                    if( !isANumber && !isABooleanValue && !isARegexPattern){
                        //checking if an underscore is present in the string
                        isUnderscorePresentInString = literalText.includes('_')

                        //checking if the string starts with certain allowed words
                        isAnAllowedString = allowedStringStartWords.some(str => literalText.startsWith(str))

                        //checking if the string is an empty string
                        isAnEmptyString = literalText.trim() === '' ? true : false

                        //checking if the string is a single symbol
                        isASingleSymbol = isTheStringASingleSymbol(literalText)
                    }

                    //checking if the string is in PascalCase
                    isAPascalCasedString = PascalCasePattern.test(literalText)
                 }

                //reporting multiword hardcoding errors
                if(isHardCoded && !isAStringInACondition && !isAStringInAllowedAttribute && !isAStringInAllowedProp 
                    && !isAStringInAttributeOfAllowedTag && !isAStringInAllowedTag && !isACSSProperty && !isAStringInThrowStatement && !isASequenceOfNumbers
                    && !isAPropertyName && !isAMemberExpression && !isAStringInNewExpression  && !isValidKeyInsideValidMethod && !isADateFormatPattern
                    && !isAStringInAllowedMethod && !isAStringInsideIgnoredBlock && !isI18NNotNeeded){
                    context.report({
                        node:node,
                        message: ERROR_MESSAGE
                    })
                }

                // reporting single word hardcoding errors
                if(isSingleWordHardcoded && !isAStringInACondition && !isAStringInAllowedAttribute && !isAStringInAllowedProp
                   && !isAStringInAttributeOfAllowedTag && !isAStringInAllowedTag && !isACSSProperty && !isAStringInThrowStatement && !isASequenceOfNumbers
                   && !isAPropertyName && !isAMemberExpression && !isAStringInNewExpression && !isValidKeyInsideValidMethod && !isADateFormatPattern
                   && !isAStringInAllowedMethod && !isAStringInsideIgnoredBlock && !isI18NNotNeeded
                   && !isSingleWordStringPassedAsArgument && !isACamelCasedString && !isUnderscorePresentInString && !isAnAllowedString
                   && !isANumber && !isABooleanValue && !isAKeyPattern && !isAStringInImportStatement && !isARegexPattern && !isAnEmptyString
                   && !isAPascalCasedString && !isASingleSymbol && !isSingleWordInitAsAssignment && !isSingleWordInitAsObjectProperty && !isSingleWordInitInArray
                   && !isAStringInExportDeclaration
                ){
                    context.report({
                        node:node,
                        message: ERROR_MESSAGE
                    })
                }
            },
            JSXElement: function(node){
                //hardcoded string pattern
                const hardcodedStringPattern = /[a-zA-Z]+/

                let isAStringInJSXText = false
                let isHardcoded = false
                let isI18NNotNeeded = false
                let ancestors = context.getAncestors();
                let programNode = ancestors.find(node => node.type === "Program")
                let comments = programNode.comments.filter(comment => commentsToIgnoreHardcodingCheck.includes(comment.value.trim().toLowerCase()))
                let validAncestors = ancestors.filter(ancestor => ancestor.type ==='JSXElement')
                let allowedTagsLocation = []
                allowedTagsLocation = validAncestors.filter(ancestor => allowedTags.includes(ancestor.openingElement.name.name))
                                
                validTagsLocationObject = allowedTagsLocation[0]
                
                //checking if JSX text is hardcoded
                let JSXWithHardcodedString = node.children.filter(child => child.type && child.type === 'JSXText' && child.value.trim() != '')
                let JSXText = ''
                if(JSXWithHardcodedString.length > 0){
                    for(let i=0; i < JSXWithHardcodedString.length; i++){
                        isAStringInJSXText = true
                        JSXText = JSXWithHardcodedString[i].value.trim()
                        isHardcoded = hardcodedStringPattern.test(JSXText)
                        //checking if valid comment is added
                        if(isHardcoded && JSXWithHardcodedString[i].loc.start.line === JSXWithHardcodedString[i].loc.end.line){
                            let hardcodingLineNumber = JSXWithHardcodedString[i].loc.start.line
                            isI18NNotNeeded = comments.some(comment => comment.loc.start.line === hardcodingLineNumber)
                        }
                        else if(isHardcoded && JSXWithHardcodedString[i].loc.start.line !== JSXWithHardcodedString[i].loc.end.line){
                            let hardcodingLineStartNumber = JSXWithHardcodedString[i].loc.start.line
                            let hardcodingLineEndNumber = JSXWithHardcodedString[i].loc.end.line
                            isI18NNotNeeded = comments.some(comment => comment.loc.start.line >= hardcodingLineStartNumber && 
                                                                        comment.loc.start.line <= hardcodingLineEndNumber)
                        }
                        if(isHardcoded && !isI18NNotNeeded){
                            let errorLineNumber = JSXWithHardcodedString[i].loc.start.line === JSXWithHardcodedString[i].loc.end.line ? JSXWithHardcodedString[i].loc.start : {line: JSXWithHardcodedString[i].loc.start.line + 1,column: JSXWithHardcodedString[i].start}
                            context.report({
                                loc:errorLineNumber,
                                message: ERROR_MESSAGE_JSX_TEXT
                            })
                        }
                    }
                }
            },
            VariableDeclarator: function(node){
                let scope = context.getScope()
                let isAllowedStringInitialized = false

                if(node.init?.type === 'Literal' && !node.init.regex){
                    let initializedString = node.init.value

                    isAllowedStringInitialized = isAnAllowedStringInitialized(initializedString)
                    if(!isAllowedStringInitialized && node.id.type !== 'ObjectPattern') {
                        stringInitializedIdentifiers.push({identifierName: node.id.name, location: node.id.loc.start, scope})
                    }
                }
                else if(node.init?.type === 'ObjectExpression'){
                    let objectPropertiesWithStringInitialized = node.init.properties.filter(property => property.value?.type === 'Literal')
                    
                    objectPropertiesWithStringInitialized.map(property => {
                        let propertyValue = property.value
                        if(!propertyValue.regex){
                            if(!isAnAllowedStringInitialized(propertyValue.value) && property.key.type === 'Identifier'){
                                stringInitObjectProperties.push({objectName:node.id.name,propertyName: property.key.name,location:property.key.loc.start,scope})
                            }
                        }  
                    })
                }
                else if(node.init?.type === 'LogicalExpression' && (node.init.left.type === 'Literal' || node.init.right.type === 'Literal')){
                    let initializedString = node.init.left.type === 'Literal' && !node.init.left.regex
                                            ? node.init.left.value
                                            : node.init.right.type === 'Literal' && !node.init.right.regex
                                            ? node.init.right.value : ''
                    isAllowedStringInitialized = isAnAllowedStringInitialized(initializedString)
                    if(!isAllowedStringInitialized){
                        stringInitializedIdentifiers.push({identifierName:node.id.name,location: node.id.loc.start, scope})
                    }    
                }
                else if(node.init?.type === 'ConditionalExpression' && (node.init.consequent.type === 'Literal' || node.init.alternate.type === 'Literal')){
                    let initializedString = node.init.consequent.type === 'Literal' && !node.init.consequent.regex
                                            ? node.init.consequent.value
                                            : node.init.alternate.type === 'Literal' && !node.init.alternate.regex
                                            ? node.init.alternate.value : ''
                    isAllowedStringInitialized = isAnAllowedStringInitialized(initializedString)
                    if(!isAllowedStringInitialized){
                        stringInitializedIdentifiers.push({identifierName:node.id.name,location: node.id.loc.start, scope})
                    }    
                }
                else if(node.id.type === 'ObjectPattern'){
                    let destructuredPropertiesInitWithString = node.id.properties.filter(property => property.type !== 'ExperimentalRestProperty' 
                                                                                        && property.value?.type === 'AssignmentPattern' 
                                                                                        && property.value.right.type === 'Literal' 
                                                                                        && !property.value.right.regex)
                    destructuredPropertiesInitWithString.map(property => {
                        if(property.value.right && !isAnAllowedStringInitialized(property.value.right.value)){
                            stringInitializedIdentifiers.push({identifierName: property.key.name,location:property.loc.start,scope})
                        }
                    })
                }

                //checking if any of the allowed method is assigned to another name and used 
                if(node.init && node.init.type === 'MemberExpression'){
                    aliasNamesOfExcludedMethods = methods.includes(node.init.property.name) ? [...aliasNamesOfExcludedMethods,node.id.name]
                                                                                            : [...aliasNamesOfExcludedMethods]
                }
                if(node.init && node.init.type === 'MemberExpression' && node.init.property.name === 'getI18NValue'){
                    if(aliasNamesOfgetI18NValueMethod.indexOf(node.id.name) === -1){
                        aliasNamesOfgetI18NValueMethod = [...aliasNamesOfgetI18NValueMethod,node.id.name]
                    }                    
                }  
            },
            AssignmentExpression: function(node){
                let scopeOfCurrentVariable = context.getScope()
                let ancestors = context.getAncestors()
                let isReassignedInIfBlock = ancestors.some(ancestor => ancestor.type === 'IfStatement')
                if(node.left?.type === 'Identifier'){
                    let variable = node.left.name
                    let variableWithStringInitialized = stringInitializedIdentifiers.find(ide => ide.identifierName === variable)

                    if(variableWithStringInitialized && !isReassignedInIfBlock &&
                        isPresentInSameScope(variableWithStringInitialized.scope,scopeOfCurrentVariable)){
                        if(node.right?.type === 'Literal'){
                            variableWithStringInitialized.location = node.loc.start
                        }
                        else if(node.right?.type === 'Identifier'){
                            variableWithStringInitialized.location = 10000
                        }
                        else if(node.right?.type === 'ConditionalExpression' && 
                            (node.right.consequent.type === 'Literal' || node.right.alternate.type === 'Literal')){
                                variableWithStringInitialized.location === node.loc.start
                        }
                        else if(node.right?.type === 'ConditionalExpression' && 
                            (node.right.consequent.type === 'Identifier' && node.right.alternate.type === 'Identifier')){
                                variableWithStringInitialized.location === 10000
                        }
                    }
                }
                if(node.left.type === 'MemberExpression' && node.right.type === 'Literal'){
                    let objectName = node.left.object.name
                    let propertyName = ''
                    if(node.left.property.type === 'Identifier'){
                        propertyName = node.left.property.name
                    }
                    else if(node.left.property.type === 'Literal'){
                        propertyName = node.left.property.name
                    }
                    let objectPropertyInitWithString = stringInitObjectProperties.find(prop => prop.objectName === objectName
                                                        && prop.propertyName === propertyName)
                    if(objectPropertyInitWithString && !isReassignedInIfBlock &&
                        isPresentInSameScope(objectPropertyInitWithString.scope,scopeOfCurrentVariable)){
                        if(node.right?.type === 'Literal'){
                            objectPropertyInitWithString.location = node.loc.start
                        }
                        else if(node.right.type === 'Identifier'){
                            objectPropertyInitWithString.location = 10000
                        }
                    }
                }
            },
            TemplateLiteral: function(node){

                const STRING_LITERAL_HARDCODING_ERROR_MESSAGE = 'Do not hardcode string literal. Use I18N key instead'
                const hardcodedStringPattern = /(\w\s)+((&|,)\s)*\w+(\s\w)*/

                //boolean variables for reporting errors
                let isStringLiteralHardCodedInAllowedTag = false 
                let isStringLiteralHardCodedInAttributeOfAllowedTag = false 
                let isStringLiteralHardCodedInCondition = false 
                let isStringLiteralHardCodedInAllowedProperty = false 
                let isStringLiteralHardCodedInAllowedProp = false 
                let isACSSProperty = false 
                let isADateFormatPattern = false 
                let isStringLiteralHardCodedInThrowStatement = false 
                let isStringLiteralHardCodedInNewExpression = false 
                let isStringLiteralHardCodedAsMemberExpression = false 
                let isValidKeyUsedInValidMethod = false 
                let isAStringLiteralInAllowedMethod = false 
                let isI18NNotNeeded = false 
                let isStringLiteralHardCodedInsideIgnoredBlock = false 
                let isSequenceOfNumsHardcodedAsStringLiteral = false 

                let ancestorsOfStringLiteral = context.getAncestors()

                let immediateAncestorOfStringLiteral = ancestorsOfStringLiteral[ancestorsOfStringLiteral.length - 1]
                let grandParentOfStringLiteral = ancestorsOfStringLiteral[ancestorsOfStringLiteral.length - 2]
                let greatGrandParentOfStringLiteral = ancestorsOfStringLiteral[ancestorsOfStringLiteral.length - 3]
                
                let isStringLiteralHardCoded = false                

                let isAnyIdentifierUsedInStringLiteral = node.expressions.length > 0 ? true : false
                let stringLiteralContentExcludingWhiteSpace = node.quasis.filter(ele => ele.value.raw.trim() !== '')
                let isOnlyStringPresentInStringLiteral = stringLiteralContentExcludingWhiteSpace.length === 1 ? true : false
                let stringLiteralStartsAt = 0
                let stringLiteralContent = ''

                if(isOnlyStringPresentInStringLiteral && !isAnyIdentifierUsedInStringLiteral){
                    stringLiteralContent = node.quasis[0].value.raw.trim()
                    stringLiteralStartsAt = node.quasis[0].loc.start.line
                    isStringLiteralHardCoded = hardcodedStringPattern.test(stringLiteralContent)
                }

                if(isStringLiteralHardCoded){
                    //checking if string literal is hardcoded in an allowed tag
                    isStringLiteralHardCodedInAllowedTag = isHardcodedInAllowedTag(stringLiteralStartsAt)

                    //checking if string literal is hardcoded in an attribute of allowed tag
                    isStringLiteralHardCodedInAttributeOfAllowedTag = isHardcodedInAttributeOfAllowedTag(ancestorsOfStringLiteral)

                    //checking if string literal is hardcoded in condition
                    isStringLiteralHardCodedInCondition = isHardcodedInCondition(immediateAncestorOfStringLiteral)

                    //checking if string literal hardcoded in an allowed key/property
                    isStringLiteralHardCodedInAllowedProperty = isHardcodedInAllowedProperty(immediateAncestorOfStringLiteral,grandParentOfStringLiteral)

                    //checking if string literal hardcoded in an allowed prop
                    if(grandParentOfStringLiteral.type === 'JSXAttribute'){
                        isStringLiteralHardCodedInAllowedProp = isHardcodedInAllowedProp(grandParentOfStringLiteral)
                    }
                    else if(immediateAncestorOfStringLiteral.type === 'LogicalExpression'
                        && grandParentOfStringLiteral.type === 'JSXExpressionContainer'
                        && greatGrandParentOfStringLiteral === 'JSXAttribute'){
                            isStringLiteralHardCodedInAllowedProp = isHardcodedInAllowedProp(greatGrandParentOfStringLiteral)
                    }

                    //checking if a CSS property is hardcoded as string literal
                    isACSSProperty = isAHardcodedCSSProperty(stringLiteralContent)

                    //checking if a date format pattern is hardcoded as string literal
                    isADateFormatPattern = isAHardcodedDateFormatPattern(stringLiteralContent)

                    //checking if string literal hardcoded in throw statement
                    isStringLiteralHardCodedInThrowStatement = isHardcodedInThrowStmt(immediateAncestorOfStringLiteral)

                    //checking if string literal hardcoded in new expression
                    isStringLiteralHardCodedInNewExpression =  isHardcodedInNewExpression(immediateAncestorOfStringLiteral)

                    //checking if string literal hardcoded as member expression - obj[`key`] - MemberExpression example
                    isStringLiteralHardCodedAsMemberExpression = isHardcodedAsMemberExpression(immediateAncestorOfStringLiteral)

                    //checking if valid key hardcoded as string literal inside valid method
                    isValidKeyUsedInValidMethod = isValidKeyInValidMethod(immediateAncestorOfStringLiteral,stringLiteralContent)

                    //checking if string literal hardcoded inside allowed methods
                    if(immediateAncestorOfStringLiteral.type === 'CallExpression'){
                        isAStringLiteralInAllowedMethod = isHardcodedInAllowedMethods(immediateAncestorOfStringLiteral)
                    }
                    else if(immediateAncestorOfStringLiteral.type === 'Property'
                        && grandParentOfStringLiteral.type === 'ObjectExpression'
                        && greatGrandParentOfStringLiteral.type === 'CallExpression'){
                            isAStringLiteralInAllowedMethod = isHardcodedInAllowedMethods(greatGrandParentOfStringLiteral)
                    }

                    //checking if a sequence of numbers are hardcoded as string literal
                    isSequenceOfNumsHardcodedAsStringLiteral = isASequenceOfNumbersHardcoded(stringLiteralContent)

                    //checking if valid comments added  to ignore i18n check
                    let commentObj = isCommentAdded(ancestorsOfStringLiteral,stringLiteralStartsAt,immediateAncestorOfStringLiteral)
                    isI18NNotNeeded = commentObj.isCommentAdded

                    //checking if valid block comments added to ignore i18n check
                    isStringLiteralHardCodedInsideIgnoredBlock = commentObj.isBlockCommentAdded
                }
                //reporting errors
                if(isStringLiteralHardCoded && !isStringLiteralHardCodedInAllowedTag && !isStringLiteralHardCodedInAttributeOfAllowedTag && !isSequenceOfNumsHardcodedAsStringLiteral
                 && !isStringLiteralHardCodedInCondition && !isStringLiteralHardCodedInAllowedProperty && !isStringLiteralHardCodedInAllowedProp
                 && !isACSSProperty && !isADateFormatPattern && !isStringLiteralHardCodedInThrowStatement && !isStringLiteralHardCodedInNewExpression
                 && !isStringLiteralHardCodedAsMemberExpression && !isValidKeyUsedInValidMethod && !isAStringLiteralInAllowedMethod
                 && !isI18NNotNeeded && !isStringLiteralHardCodedInsideIgnoredBlock
                ){
                    context.report({
                        node: node,
                        message: STRING_LITERAL_HARDCODING_ERROR_MESSAGE
                    })
                }
            },
            JSXAttribute: function(node){
                let stringInitialisedVariablePassedAsProp
                let stringInitializedObjectPropertyPassedAsProp
                let stringAssignedObjectProps = []
                let commentObjForVar = {
                    isCommentAdded: false,
                    isBlockCommentAdded: false
                }
                let commentObjForObjProp = {
                    isCommentAdded: false,
                    isBlockCommentAdded: false
                }

                let ancestors = context.getAncestors()
                let currentScope = context.getScope()
                if( !isHardcodedInAllowedProp(node)){
                    if(node.value !== null && node.value.type === 'JSXExpressionContainer' 
                        && node.name.type === 'JSXIdentifier'){                    
                        
                        let stringAssignedIdentifiers = stringInitializedIdentifiers.filter(identifier => 
                                                        identifier.identifierName === node.value.expression.name
                                                        || identifier.identifierName === node.value.expression?.consequent?.name
                                                        || identifier.identifierName === node.value.expression?.alternate?.name
                                                    )
                        stringAssignedIdentifiers.map(
                            element => {
                                let distance = 5000
                                if(node.loc.start.line - element.location.line < distance){
                                    distance = node.loc.start.line - element.location.line
                                    stringInitialisedVariablePassedAsProp = element
                                } 
                            }
                        )
                        if(stringAssignedIdentifiers.length === 0 && node.value.expression.type === 'MemberExpression'){
                            stringAssignedObjectProps = stringInitObjectProperties.filter(objectProp => objectProp.objectName === node.value.expression?.object.name
                                                            && objectProp.propertyName === node.value.expression?.property.name)
                            
                            stringAssignedObjectProps.map(
                                element => {
                                    let distance = 5000
                                    if(node.loc.start.line - element.location.line < distance){
                                        distance = element.location.line
                                        stringInitializedObjectPropertyPassedAsProp = element
                                    }
                                }
                            )
                        }
                        if(stringInitialisedVariablePassedAsProp){
                            commentObjForVar = isCommentAdded(ancestors,stringInitialisedVariablePassedAsProp.location.line)
                        }
                        if(stringInitializedObjectPropertyPassedAsProp){
                            commentObjForObjProp = isCommentAdded(ancestors,stringInitializedObjectPropertyPassedAsProp.location.line)
                        }

                        //checking if present in same scope
                        let isStringInitVariableInSameScope = isPresentInSameScope(stringInitialisedVariablePassedAsProp?.scope,currentScope)
                        let isObjectPropInSameScope = isPresentInSameScope(stringInitializedObjectPropertyPassedAsProp?.scope,currentScope)
                        
                        //reporting the errors
                        if(stringInitialisedVariablePassedAsProp && isStringInitVariableInSameScope
                            && !commentObjForVar.isCommentAdded && !commentObjForVar.isBlockCommentAdded
                        ){
                            context.report({
                                loc: stringInitialisedVariablePassedAsProp.location,
                                message: `The variable "${stringInitialisedVariablePassedAsProp.identifierName}" is assigned a string and passed as prop value to the prop "${node.name.name}" in line ${node.loc.start.line} which may cause I18N issue. So use appropriate I18N key instead of assigning a string`
                            })
                        }
                        else if(stringInitializedObjectPropertyPassedAsProp && isObjectPropInSameScope
                            && !commentObjForObjProp.isCommentAdded && !commentObjForObjProp.isBlockCommentAdded
                        ){
                            context.report({
                                loc: stringInitializedObjectPropertyPassedAsProp.location,
                                message: `The property "${stringInitializedObjectPropertyPassedAsProp.propertyName}" of object ${stringInitializedObjectPropertyPassedAsProp.objectName} is initialized with string and passed as prop to the prop "${node.name.name}" in line ${node.loc.start.line} which may cause I18N issue. So use appropriate I18N key instead of assigning a string`
                            })
                        }
                    }
                }
            }
        }
    }
}

function isHardcodedInAllowedTag(stringStartsAt){
    if(validTagsLocationObject){
        if((validTagsLocationObject.loc.start.line <= stringStartsAt)
            && (stringStartsAt <= validTagsLocationObject.loc.end.line)){
                return true
            }
    }
}

function isHardcodedInAttributeOfAllowedTag(ancestors){
    let allowedTagAncestor = ancestors.find(ancestor => ancestor && ancestor.type === 'JSXOpeningElement')
    return allowedTags.some(tag => allowedTagAncestor && tag === allowedTagAncestor.name.name)
}

function isHardcodedInCondition(immediateAncestor){
    if(immediateAncestor.type === 'BinaryExpression' ||
       immediateAncestor.type === 'ConditionalExpression' ||
       immediateAncestor.type === 'SwitchCase'){
        return true
    }
    else{
        return false
    }
}

function isHardcodedInAllowedProperty(immediateAncestor,grandParent){
    let isAStringInAllowedProperty = false
    if(immediateAncestor.type === 'Property'){
        isAStringInAllowedProperty = objectKeys.some(objectKey => objectKey===immediateAncestor.key.name)
        if( !isAStringInAllowedProperty ){
            isAStringInAllowedProperty = allowedPropStartingWords.some(
                allowedPropStartWord => immediateAncestor.key.name && immediateAncestor.key.name.startsWith(allowedPropStartWord)
            )
        }
        if( !isAStringInAllowedProperty ){
            isAStringInAllowedProperty = allowedPropEndingWords.some(
                allowedPropEndWord => immediateAncestor.key.name && immediateAncestor.key.name.endsWith(allowedPropEndWord)
            )
        }
    }
    else if(grandParent.type === 'Property'){
        isAStringInAllowedProperty = objectKeys.some(objectKey => objectKey===grandParent.key.name)
        if( !isAStringInAllowedProperty ){
            isAStringInAllowedProperty = allowedPropStartingWords.some(
                allowedPropStartWord => grandParent.key.name && grandParent.key.name.startsWith(allowedPropStartWord)
            )
        }
        if( !isAStringInAllowedProperty ){
            isAStringInAllowedProperty = allowedPropEndingWords.some(
                allowedPropEndWord => grandParent.key.name && grandParent.key.name.endsWith(allowedPropEndWord)
            )
        }
    }
    return isAStringInAllowedProperty
}

function isHardcodedInAllowedProp(ancestor){
    let isHardcodedInAllowedProp = false
    
    isHardcodedInAllowedProp = props.some(props => props === ancestor.name.name)
    if( !isHardcodedInAllowedProp ){
        isHardcodedInAllowedProp = allowedPropStartingWords.some(startWord => ancestor.name.name.startsWith(startWord))
        if( !isHardcodedInAllowedProp ){
            isHardcodedInAllowedProp = allowedPropEndingWords.some(endWord => ancestor.name.name.endsWith(endWord))
        }
        if( !isHardcodedInAllowedProp ){
            isHardcodedInAllowedProp = allowedWordsThatCanBePresentInAComponentNameToExcludeTheComponent.some(
                allowedName => ancestor.name.name.includes(allowedName)
            )
        }
    }
    return isHardcodedInAllowedProp
}

function isAHardcodedCSSProperty(str){
    return cssStylingPattern.test(str)
}

function isAHardcodedDateFormatPattern(str){
    const dateFormatPatternsRegExp = dateFormatPatterns.map((pattern) => {
        return new RegExp(pattern)
    })
    return dateFormatPatternsRegExp.some(pattern => pattern.test(str))
}

function isHardcodedInThrowStmt(immediateAncestor){
    return immediateAncestor.type === 'ThrowStatement'
}

function isHardcodedInNewExpression(immediateAncestor){
    return immediateAncestor.type === 'NewExpression'
}

function isHardCodedInObjectKey(immediateAncestor){
    return immediateAncestor.type=== 'Property' && immediateAncestor.key.type === 'Literal'
}

function isHardcodedAsMemberExpression(immediateAncestor){
    return immediateAncestor.type === 'MemberExpression'&& 
            (immediateAncestor.property.type === 'Literal' || immediateAncestor.property.type === 'TemplateLiteral')
}

function isValidKeyInValidMethod(immediateAncestor,stringContent){
    let isAValidKey = keys.includes(stringContent)
    let isValidKeyUsedInValidMethod = false
    
    if(isAValidKey){
        if(immediateAncestor.type === 'CallExpression' && immediateAncestor.callee){
            if(immediateAncestor.callee.type === 'MemberExpression'){
                if(immediateAncestor.callee.property.name === 'getI18NValue'){
                    isValidKeyUsedInValidMethod = true
                }
                else if(!isValidKeyUsedInValidMethod){
                    isValidKeyUsedInValidMethod = aliasNamesOfgetI18NValueMethod.some(method => method === immediateAncestor.callee.property.name)
                }
            }
            else if(immediateAncestor.callee.type === 'Identifier'){
                if(immediateAncestor.callee.name === 'getI18NValue'){
                    isValidKeyUsedInValidMethod = true
                }
                else if(!isValidKeyUsedInValidMethod){
                    isValidKeyUsedInValidMethod = aliasNamesOfgetI18NValueMethod.some(method => method === immediateAncestor.callee.name)
                }
            }
        }
    }
    return isValidKeyUsedInValidMethod
}

function isHardcodedInAllowedMethods(ancestor){
    let isAStringInAllowedMethod = false
    
    if(ancestor.callee && ancestor.callee.type ==='MemberExpression'){
        isAStringInAllowedMethod = methods.some(method => method === ancestor.callee.property.name)
        if(!isAStringInAllowedMethod){
            isAStringInAllowedMethod = aliasNamesOfExcludedMethods.some(method => method === ancestor.callee.property.name)
        }
    }
    else if(ancestor.callee && ancestor.callee.type === 'Identifier'){
        isAStringInAllowedMethod = methods.some(method => method === ancestor.callee.name)
        if(!isAStringInAllowedMethod){
            isAStringInAllowedMethod = aliasNamesOfExcludedMethods.some(method => method === ancestor.callee.name)
        }
    }  
    return isAStringInAllowedMethod
}

function isCommentAdded(ancestors,stringStartsAt,immediateAncestor = undefined){
    let commentObj = {
        isCommentAdded: false,
        isBlockCommentAdded: false
    }
    let programNode = ancestors.find(ancestor => ancestor.type === 'Program')
    let comments = programNode.comments.filter(Line => commentsToIgnoreHardcodingCheck.includes(Line.value.trim().toLowerCase()))
    let blockCommentsStart = programNode.comments.filter(Line => stringsAllowedAsBlockCommentStart.includes(Line.value.trim().toLowerCase()))
    let blockCommentsEnd = programNode.comments.filter(Line => stringsAllowedAsBlockCommentEnd.includes(Line.value.trim().toLowerCase()))

    commentObj.isCommentAdded = comments.some(comment => comment.loc.start.line === stringStartsAt
                                                        || stringStartsAt - comment.loc.start.line === 1)

    if(!commentObj.isCommentAdded && immediateAncestor?.type === 'CallExpression'){
        commentObj.isCommentAdded = comments.some(comment => comment.loc.start.line === immediateAncestor?.loc.start.line
                                                        || (immediateAncestor?.loc.start.line < comment.loc.start.line && 
                                                            comment.loc.start.line < immediateAncestor?.loc.end.line)
                                                        || comment.loc.start.line === immediateAncestor?.loc.end.line)
    }

    if(blockCommentsStart.length === blockCommentsEnd.length){
        for(let i=0;i<blockCommentsStart.length;i++){
            if(blockCommentsStart[i].loc.start.line <= stringStartsAt
                &&stringStartsAt<= blockCommentsEnd[i].loc.start.line){
                    commentObj.isBlockCommentAdded = true
                }
            if(commentObj.isBlockCommentAdded){
                break;
            }
        }
    }
    return commentObj
}

function isASequenceOfNumbersHardcoded(str){
    return sequenceOfNumbers.test(str)
}

function isTheStringASingleSymbol(str){

    if(str.trim().length === 1){
        return symbolPattern.test(str)
    }
    else{
        return false
    }
}

function isPresentInSameScope(scopeWhenVariableWasInitWithString, scopeWhenVariableWasReassigned){
    let scope = scopeWhenVariableWasReassigned
    while(scope){
        if(scope === scopeWhenVariableWasInitWithString){
            return true
        }
        scope = scope.upper
    }
    return false
}

function isAnAllowedStringInitialized(initializedString){
    let isAValidString = false
    
    isAValidString = numberPattern.test(initializedString) || booleanPattern.test(initializedString)
                    || initializedString === null
    if(!isAValidString){
        isAValidString = keyPattern.test(initializedString) || camelCasePattern.test(initializedString) || initializedString.includes('_') ||
                         initializedString.trim() === '' || PascalCasePattern.test(initializedString) ||
                         isTheStringASingleSymbol(initializedString) || allowedStringStartWords.some(str => initializedString.startsWith(str))
    }
    return isAValidString
}